// RPi ADC measurement function for IC Checker Project.
// See article on Transistor Gijutsu 2023/3. 
// 
// 2022/12 Created, by Hiroyuki Yamada
// 
#include <stdio.h>
#include <string.h>

#include <wiringPi.h>
#include <wiringPiI2C.h>
#include <wiringPiSPI.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>

int g_i2cFD[3];

/////////////////////////////////////////
// Constant Definition
#define ADC_VREF  (3.3*30/20) // includes pre-divider
#define VCC   3.3
#define ADC_RESOLUTION 1024
#define RSRC1_RES (2.7+4.1)
#define RSRC1_PIN 24

#define RSRC2_RES (30+4.1)
#define RSRC2_PIN 25

#define RSRC3_RES (68+4.1)
#define RSRC3_PIN 26

#define RSRC4_RES (150+4.1)
#define RSRC4_PIN 27

#define LED2_PIN 13
#define CLAMP_PIN 16

#define NC_THRESHOLD 1.5f // 
#define R_THRESHOLD  0.6f // Resistance is calculated if voltage is below this value

#define SHORT_THRESHOLD 200

#define DEBUG_MODE 0 

/////////////////////////////////////////
// Basic function

int GetADCValue(int ch)
{
	// MCP3002
	uint8_t d[2];
	d[0] = 0b11010000;
	if(ch) d[0] |= 0b00100000;
	d[1] = 0;
	wiringPiSPIDataRW(0, d, 2);
	//printf("Raw: %d,%d,%d\n", d[0], d[1], d[2]);
	return ((((int)d[0] << 8) | d[1])/2) & 0x03FF;
}

void Expander_Init()
{
	// set all pins hi-z
	wiringPiI2CWriteReg8(g_i2cFD[0], 6, 0xFF);
	wiringPiI2CWriteReg8(g_i2cFD[0], 7, 0xFF);
	wiringPiI2CWriteReg8(g_i2cFD[1], 6, 0xFF);
	wiringPiI2CWriteReg8(g_i2cFD[1], 7, 0xFF);
	wiringPiI2CWriteReg8(g_i2cFD[2], 6, 0xFF);
	
	// MPX Sel: Hi-Z
	wiringPiI2CWriteReg8(g_i2cFD[2], 3, 0x00);
	wiringPiI2CWriteReg8(g_i2cFD[2], 7, 0xFF);
}

// Multiplexer
#define MPX_1 0x1E // INH
#define MPX_2 0x1D
#define MPX_3 0x1B
#define MPX_4 0x17
#define MPX_5 0x0F
void SetMpxChannel(int ch)
{
	const uint8_t mpxList[40] = 
	{
		// bit[4:0]: MPX Inhibit(1)
		// bit[7:5]: MPX Common Select Signal
		MPX_1|(0<<5),MPX_1|(1<<5),MPX_1|(2<<5),MPX_1|(3<<5),
		MPX_2|(0<<5),MPX_2|(1<<5),MPX_2|(2<<5),MPX_2|(3<<5),
		MPX_3|(0<<5),MPX_3|(1<<5),MPX_3|(2<<5),MPX_3|(3<<5),
		MPX_4|(0<<5),MPX_4|(1<<5),MPX_4|(2<<5),MPX_4|(3<<5),
		MPX_5|(0<<5),MPX_5|(1<<5),MPX_5|(2<<5),MPX_5|(3<<5),
		
		MPX_5|(4<<5),MPX_5|(5<<5),MPX_5|(6<<5),MPX_5|(7<<5),
		MPX_4|(4<<5),MPX_4|(5<<5),MPX_4|(6<<5),MPX_4|(7<<5),
		MPX_3|(4<<5),MPX_3|(5<<5),MPX_3|(6<<5),MPX_3|(7<<5),
		MPX_2|(4<<5),MPX_2|(5<<5),MPX_2|(6<<5),MPX_2|(7<<5),
		MPX_1|(4<<5),MPX_1|(5<<5),MPX_1|(6<<5),MPX_1|(7<<5)
	};
	
	if(ch<=0 || ch>=41)
	{
		// All Inhibit
		wiringPiI2CWriteReg8(g_i2cFD[2], 3, 0);
		wiringPiI2CWriteReg8(g_i2cFD[2], 7, 0xFF);
		return;
	}

	// 1: Hi-Z (Reg 7: 1), 0: Output Low (Reg 7: 0)
	wiringPiI2CWriteReg8(g_i2cFD[2], 3, 0);
	wiringPiI2CWriteReg8(g_i2cFD[2], 7, mpxList[ch-1]);
}

/////////////////////////////////////////

void SetPin(int pin_num, int state)
{
	if(pin_num<=0 || pin_num>=41)
	{
		return;
	}
	
	// state: 0(Hi-Z),1(L),2(H)
	
	const uint8_t fdList[5]  = {0,1,2, 1,0};
	const uint8_t regList[5] = {2,2,2, 3,3}; // Output port reg (+4 to config reg)
	int b = (pin_num-1) % 8; // bit
	int i = (pin_num-1)/8;
	
	if(state == 1) // Out L
	{
		// PORT
		int val = wiringPiI2CReadReg8(g_i2cFD[fdList[i]], regList[i]);
		val = val & ~(1<<b); // set L
		wiringPiI2CWriteReg8(g_i2cFD[fdList[i]], regList[i], val);

		// DIR
		val = wiringPiI2CReadReg8(g_i2cFD[fdList[i]], regList[i]+4);
		val = val & ~(1<<b); // output
		wiringPiI2CWriteReg8(g_i2cFD[fdList[i]], regList[i]+4, val);
		
	}
	else if(state == 2) // Out H
	{
		// PORT
		int val = wiringPiI2CReadReg8(g_i2cFD[fdList[i]], regList[i]);
		val = val | (1<<b); // set H
		wiringPiI2CWriteReg8(g_i2cFD[fdList[i]], regList[i], val);

		// DIR
		val = wiringPiI2CReadReg8(g_i2cFD[fdList[i]], regList[i]+4);
		val = val & ~(1<<b); // output
		wiringPiI2CWriteReg8(g_i2cFD[fdList[i]], regList[i]+4, val);
	}
	else // Hi-Z
	{
		// DIR
		// Read current state
		int val = wiringPiI2CReadReg8(g_i2cFD[fdList[i]], regList[i]+4);
		val = val | (1<<b); // input
		wiringPiI2CWriteReg8(g_i2cFD[fdList[i]], regList[i]+4, val );
	}
}
void SetAllPin(int state)
{
	if(state == 1) // Out L
	{
		// PORT
		wiringPiI2CWriteReg8(g_i2cFD[0], 2, 0);
		wiringPiI2CWriteReg8(g_i2cFD[0], 3, 0);
		wiringPiI2CWriteReg8(g_i2cFD[1], 2, 0);
		wiringPiI2CWriteReg8(g_i2cFD[1], 3, 0);
		wiringPiI2CWriteReg8(g_i2cFD[2], 2, 0);
		
		// DIR
		wiringPiI2CWriteReg8(g_i2cFD[0], 6, 0);
		wiringPiI2CWriteReg8(g_i2cFD[0], 7, 0);
		wiringPiI2CWriteReg8(g_i2cFD[1], 6, 0);
		wiringPiI2CWriteReg8(g_i2cFD[1], 7, 0);
		wiringPiI2CWriteReg8(g_i2cFD[2], 6, 0);
	}
	else if(state == 2) // Out H
	{
		// PORT
		wiringPiI2CWriteReg8(g_i2cFD[0], 2, 0xFF);
		wiringPiI2CWriteReg8(g_i2cFD[0], 3, 0xFF);
		wiringPiI2CWriteReg8(g_i2cFD[1], 2, 0xFF);
		wiringPiI2CWriteReg8(g_i2cFD[1], 3, 0xFF);
		wiringPiI2CWriteReg8(g_i2cFD[2], 2, 0xFF);
		
		// DIR
		wiringPiI2CWriteReg8(g_i2cFD[0], 6, 0);
		wiringPiI2CWriteReg8(g_i2cFD[0], 7, 0);
		wiringPiI2CWriteReg8(g_i2cFD[1], 6, 0);
		wiringPiI2CWriteReg8(g_i2cFD[1], 7, 0);
		wiringPiI2CWriteReg8(g_i2cFD[2], 6, 0);
		
	}
	else // Hi-Z
	{
		// DIR
		wiringPiI2CWriteReg8(g_i2cFD[0], 6, 0xFF);
		wiringPiI2CWriteReg8(g_i2cFD[0], 7, 0xFF);
		wiringPiI2CWriteReg8(g_i2cFD[1], 6, 0xFF);
		wiringPiI2CWriteReg8(g_i2cFD[1], 7, 0xFF);
		wiringPiI2CWriteReg8(g_i2cFD[2], 6, 0xFF);
	}
}

int GetPinVoltage(int pin_num)
{
	SetMpxChannel(pin_num);
	usleep(3000);
	return GetADCValue(0);
}

void PrintAllPinState()
{
	int i;
	const uint8_t fdList[5]  = {0,1,2, 1,0};
	const uint8_t regList[5] = {0,0,0, 1,1};
	for(i=0; i<5; i++)
	{
		int val = wiringPiI2CReadReg8(g_i2cFD[fdList[i]], regList[i]);
		// print val
		int b;
		for(b=0; b<8; b++){
			if((0x80 >> b) & val) putchar('1');
			else putchar('0');
		}
		putchar(' ');
	}
	printf("\n");
}
void PrintAllPinVoltage()
{
	pullUpDnControl(RSRC1_PIN, PUD_OFF); // Required.
	pullUpDnControl(RSRC2_PIN, PUD_OFF); 
	pullUpDnControl(RSRC3_PIN, PUD_OFF); 
	pullUpDnControl(RSRC4_PIN, PUD_OFF); 
	pinMode (RSRC1_PIN, INPUT);
	pinMode (RSRC2_PIN, INPUT);
	pinMode (RSRC3_PIN, INPUT);
	pinMode (RSRC4_PIN, INPUT);
	
	int i;
	for(i=0; i<40; i++)
	{
		
		int voltage = GetPinVoltage(i+1);
		printf("P%d,%.3f,%d\n", i+1, voltage*ADC_VREF/1024, voltage);
		//pinMode (24, INPUT);
	}
}
/////////////////////////////////////////

#define PIN_IS_DIODE	0x00000000
#define PIN_IS_RESISTOR	0x10000000
#define PIN_IS_NC		0x20000000
#define PIN_IS_SHORTED	0x30000000
#define PIN_TYPE_MASK	0x70000000
const int   PIN_DEF[4] = {RSRC1_PIN, RSRC2_PIN, RSRC3_PIN, RSRC4_PIN};
const float RES_DEF[4] = {RSRC1_RES, RSRC2_RES, RSRC3_RES, RSRC4_RES};


void DiodeTest_PrintResult(int result)
{
	//printf("RAW=%d, ", raw_result[i*pins+j]);
	// pin i: GND, pin j: injection pin
	switch(result & PIN_TYPE_MASK)
	{
	case PIN_IS_RESISTOR:
		printf("Resistor,%d\n", (result & 0xFFFFFF) );
		break;
	case PIN_IS_NC:
		printf("NC,\n");
		break;
	case PIN_IS_SHORTED:
		printf("Shorted,0\n");
		break;
	default:
		printf("Diode,%0.3f\n", (result & 0x3FFF) * ADC_VREF / ADC_RESOLUTION );
		break;
	}
}
int DiodeTest_Analyze(int raw_result[4])
{
	int cond;
	int result = 0;
	int impedance[4];
	float voltage[4];
	for(cond=0; cond<4; cond++)
	{
		voltage[cond] = (raw_result[cond] * ADC_VREF / ADC_RESOLUTION);
		// VCC---[R1]---[Z]---GND
		// Z = V/I
		// I = VCC/(R1+Z)
		// -> VCC*Z = V*(R1+Z) -> Z=V*R1/(VCC-V)
		if(voltage[cond] > NC_THRESHOLD) // NC
		{
			// NC
			impedance[cond] = -1;
		}
		else
		{
			impedance[cond] = voltage[cond]*RES_DEF[cond]/(VCC-voltage[cond]) * 1000; // kohm
		}
	}
	if(DEBUG_MODE)
	{
		printf("Voltage: %.3f,%.3f,%.3f,%.3f\n", voltage[0], voltage[1], voltage[2], voltage[3]);
		printf("Impedance: %d,%d,%d,%d\n", impedance[0], impedance[1], impedance[2], impedance[3]);
	}
	// Decision
	if( impedance[0] < 0 || impedance[3] < 0) // too much resistance
	{
		result = PIN_IS_NC;
	}
	else if( impedance[0] < SHORT_THRESHOLD) // too small resistance
	{
		result = PIN_IS_SHORTED;
	}
	else
	{
		const float rx_voltage_max = R_THRESHOLD;
		// Resistor?
		int isResistor = 0;
		if(voltage[0] > NC_THRESHOLD) isResistor = 1;
		for(cond=0; cond<3; cond++)
		{
			if((voltage[cond] < rx_voltage_max) && 
				(abs(impedance[cond] - impedance[cond+1]) < impedance[cond]*0.1) )
			{
			  	isResistor = 1; // impedance is same 
			}
		}
		if( isResistor)
		{
			// Get adequate range
			if(voltage[0] <= rx_voltage_max)
			{
				result = PIN_IS_RESISTOR | (int)(impedance[0]);
			}
			else if(voltage[1] <= rx_voltage_max)
			{
				result = PIN_IS_RESISTOR | (int)(impedance[1]);
			}
			else if(voltage[2] <= rx_voltage_max)
			{
				result = PIN_IS_RESISTOR | (int)(impedance[2]);
			}
			else // too high
			{
				result = PIN_IS_RESISTOR | (int)(impedance[3]);
			}
		}
		else // impedance is different
		{
			result = PIN_IS_DIODE | (raw_result[0]);
		}
	}
	return result;
}

// pin number is one of 40P socket
void DiodeTestSinglePin(int pin_anode, int pin_kathode)
{
	int raw_result[4], result;
	int cond, i;
	
	// Measure
	for(cond=0; cond<4; cond++)
	{
		for(i=0; i<4; i++)
		{
			pullUpDnControl(PIN_DEF[i], PUD_OFF); // Required.
			pinMode (PIN_DEF[i], INPUT);
		}
		pinMode (PIN_DEF[cond], OUTPUT);
		digitalWrite(PIN_DEF[cond], HIGH);
		usleep(1000);
		
		SetPin(pin_kathode, 1); // Out L
		// pin i: GND, pin j: injection pin
		raw_result[cond] = GetPinVoltage(pin_anode);
		SetPin(pin_kathode, 0); // High-Z
		usleep(1000);
	}
	// Reset GPIO
	for(i=0; i<4; i++)
	{
		pullUpDnControl(PIN_DEF[i], PUD_OFF); // Required.
		pinMode (PIN_DEF[i], INPUT);
	}
	SetMpxChannel(0);
	
	printf("Raw Result: %d,%d,%d,%d\n", 
		raw_result[0], raw_result[1], raw_result[2], raw_result[3]);
	
	result = DiodeTest_Analyze(raw_result);
	DiodeTest_PrintResult(result);
}

/////////////////////////////////////////
void DiodeTestAndPrint(int pins)
{
	int l = 4 * sizeof(int) * pins*pins;
	int* raw_result = (int*)malloc(4*l); // 40P x 40P x 4
	memset(raw_result, 0, 4*l);
	int* result = (int*)malloc(l); // 40P x 40P
	memset(result, 0, l);
	
	printf("Pin Count,%d\n", pins);
	// Measure
	int cond;
	int i, j;
	int pin_min = 1 + (20-pins/2);
	for(cond=0; cond<4; cond++)
	{
		for(i=0; i<4; i++)
		{
			pullUpDnControl(PIN_DEF[i], PUD_OFF); // Required.
			pinMode (PIN_DEF[i], INPUT);
		}
		pinMode (PIN_DEF[cond], OUTPUT);
		digitalWrite(PIN_DEF[cond], HIGH);
		usleep(1000);
		
		for(i=0; i<pins; i++)
		{
			SetAllPin(1); // turn All pin L to discharge
			usleep(100);
			SetAllPin(0); // turn All pin Hi-Z
			
			SetPin(pin_min+i, 1); // Out L
			for(j=0; j<pins; j++)
			{
				if(i==j) continue;
				
				// pin i: GND, pin j: injection pin
				raw_result[cond*pins*pins + i*pins + j] = GetPinVoltage(pin_min+j);
			}
			SetPin(pin_min+i, 0); // High-Z
			SetMpxChannel(0);
		}
	}
	// Reset GPIO (24-27)
	for(i=0; i<4; i++)
	{
		pullUpDnControl(PIN_DEF[i], PUD_OFF); // Required.
		pinMode (PIN_DEF[i], INPUT);
	}
	
	// Analyze
	for(i=0; i<pins; i++)
	{
		for(j=0; j<pins; j++)
		{
			int rr[4];
			for(cond=0; cond<4; cond++)
			{
				rr[cond] = raw_result[cond*pins*pins + i*pins + j];
			}
			result[i*pins+j] = DiodeTest_Analyze(rr);
		}
	}
	// Compensate resistor value (avoid measure under reverse biased)
	for(i=0; i<pins; i++)
	{
		for(j=0; j<pins; j++)
		{
			if(
				((result[i*pins+j] & PIN_TYPE_MASK) == PIN_IS_RESISTOR) && 
				((result[j*pins+i] & PIN_TYPE_MASK) == PIN_IS_RESISTOR))
			{
				// Apply higher one
				if(result[i*pins+j] > result[j*pins+i])
				{
					result[j*pins+i] = result[i*pins+j];
				}
				else
				{
					result[i*pins+j] = result[j*pins+i];
				}
			}
		}
	}
	
	// Estimate GND pin
	int gnd_pin = 0, vcc_pin = 0;
	int gnd_score_max = 0, vcc_score_max;
	for(i=0; i<pins; i++)
	{
		// For gnd
		int score = 0;
		for(j=0; j<pins; j++)
		{
			// In this time, i: inject, j: gnd
			int r = result[j*pins+i];
			if((r & PIN_TYPE_MASK) == PIN_IS_DIODE)
			{
				score += 0x10000 | (0xFFFF - (r & 0x3FFF)); // VF
			}
		}
		if(score > gnd_score_max)
		{
			gnd_score_max = score;
			gnd_pin = 1+i;
		}
		// For vcc
		score = 0;
		for(j=0; j<pins; j++)
		{
			// In this time, j: inject, i: gnd
			int r = result[i*pins+j];
			if((r & PIN_TYPE_MASK) == PIN_IS_DIODE)
			{
				score += 0x10000 | (0xFFFF - (r & 0x3FFF)); // VF
			}
		}
		if(score > vcc_score_max)
		{
			vcc_score_max = score;
			vcc_pin = 1+i;
		}
	}
	printf("GND,%d\n", gnd_pin);
	printf("VCC,%d\n", vcc_pin);
	
	// Print all results
	for(i=0; i<pins; i++)
	{
		for(j=0; j<pins; j++)
		{
			DiodeTest_PrintResult(result[j*pins + i]);
		}
	}
	
	free(result);
}

int main(int argc, char** argv)
{
	if(argc <= 1){
		printf("%s -diode_test [pin count]\n", argv[0]);
		printf("%s -diode_test [pin_num_anode] [pin_num_kathode]\n", argv[0]);
		printf("%s -set_pin [pin number(1-40)] [h or l or z]\n", argv[0]);
		printf("%s -get_all_pin\n", argv[0]);
		printf("%s -reset_all_pin\n", argv[0]);
		printf("%s -get_all_pin_voltage\n", argv[0]);
		return 0;
	}
	
	// Initialize
	wiringPiSetupGpio ();
	wiringPiSPISetup (0, 1000000); // SPI
	g_i2cFD[0] = wiringPiI2CSetup(0x21); // I2C
	g_i2cFD[1] = wiringPiI2CSetup(0x22);
	g_i2cFD[2] = wiringPiI2CSetup(0x23);
	if((g_i2cFD[0]<0) || (g_i2cFD[1]<0) || (g_i2cFD[2] < 0))
	{
		printf("Error: I2C Failed.\n");
		return -1;
	}
	
	// Status LED
	pinMode (LED2_PIN, OUTPUT);
	digitalWrite(LED2_PIN, HIGH);
	
	if(strcasecmp(argv[1], "-led_test") == 0)
	{
		pinMode (13, OUTPUT);
		digitalWrite(13, HIGH);
		sleep(1);
		pinMode (13, INPUT);
	}
	
	if(strcasecmp(argv[1], "-diode_test") == 0)
	{
		// Clamp Enable
		pinMode (CLAMP_PIN, OUTPUT);
		digitalWrite(CLAMP_PIN, HIGH);
		
		if(argc == 3)
		{
			int pin_cnt = atoi(argv[2]);
			if(pin_cnt<=0) pin_cnt = 1;
			if(pin_cnt>40) pin_cnt = 40;
			DiodeTestAndPrint(pin_cnt);
		}
		else if(argc == 4)
		{
			int pin1 = atoi(argv[2]);
			int pin2 = atoi(argv[3]);
			
			DiodeTestSinglePin(pin1, pin2);
		}
		
		// Clamp Disable
		pinMode (CLAMP_PIN, OUTPUT);
		digitalWrite(CLAMP_PIN, LOW);
	}
	if(strcasecmp(argv[1], "-set_pin") == 0 && (argc == 4))
	{
		int pin = atoi(argv[2]);
		if(pin<=0) pin = 1;
		if(pin>40) pin = 40;
		int state = 0;
		if(argv[3][0] == 'H' || argv[3][0] == 'h'){
			state = 2;
		}
		if(argv[3][0] == 'L' || argv[3][0] == 'l'){
			state = 1;
		}
		SetPin(pin, state);
	}


	if(strcasecmp(argv[1], "-get_all_pin") == 0)
	{
		PrintAllPinState();
	}
	if(strcasecmp(argv[1], "-get_all_pin_voltage") == 0)
	{
		PrintAllPinVoltage();
	}
	if(strcasecmp(argv[1], "-reset_all_pin") == 0)
	{
		// Clamp Disable
		pinMode (CLAMP_PIN, OUTPUT);
		digitalWrite(CLAMP_PIN, LOW);
		Expander_Init();
	}
	digitalWrite(LED2_PIN, LOW);

	return 0;
}
